/* 
 * Li-ion battery solar charger for wireless weather station
 * File:   main.c
 * Author: Hirofumi Momose (hirofumi[at]mmomose.com)
 *
 * Created on 2016/09/20, 22:31
 */

#define _XTAL_FREQ  16000000
#define EEPROM_SIZE 256
#include <xc.h>

/** CONFIGURATION **************************************************/
// CONFIG1
#pragma config CPD      = OFF
#pragma config IESO     = OFF
#pragma config FOSC     = INTOSC
#pragma config FCMEN    = OFF
#pragma config MCLRE    = ON
#pragma config WDTE     = SWDTEN // SWDTEN ON OFF NSLEEP
#pragma config CP       = OFF
#pragma config PWRTE    = OFF
#pragma config CLKOUTEN = OFF
#pragma config BOREN    = ON

// CONFIG2
#pragma config WRT      = OFF
#pragma config PLLEN    = OFF
#pragma config STVREN   = OFF
#pragma config LVP      = OFF
#pragma config BORV     = LO    // Trip point LO:1.9V HI:2.5V

#define ADCH_VBAT   (9)
#define ADCH_IBAT   (10)
#define ADCH_VPV    (2)
#define ADLSB_VBAT  (0.005F)    // VBAT 5mV/LSB
#define ADLSB_IBAT  (0.002F)    // IBAT 2mA/LSB
#define ADLSB_VPV   (0.0202F)   // VPV 20.2mV/LSB

#define LVPOFF      (2.8F)      // Low batt cut-off voltage
#define LVPON       (3.0F)      // Low batt restore voltage
#define VREADY      (6.0F)      // Operation threshold
#define VLIM_MAX    (4.2F)      // CV limit
#define ILIM_MAX    (1.0F)      // CC limit

#define PVGAIN      (0.50F)
#define IVGAIN      (1.00F)
#define PIGAIN      (0.25F)
#define IIGAIN      (0.50F)
#define PMGAIN      (0.10F)
#define IMGAIN      (0.20F)

#define I2C_ADDR    (0x40)      // I2C my address
#define I2CLSB_VBAT (0.02F)     // VBAT 20mV/LSB
#define I2CLSB_IBAT (0.01F)     // IBAT 10mA/LSB
#define I2CLSB_VPV  (0.10F)     // VPV 100mV/LSB

#define CHG_MASK    (0x03)      // Mode mask
#define CHG_OFF     (0x00)      //  Not charging
#define CHG_CV      (0x01)      //  CV mode
#define CHG_CC      (0x02)      //  CC mode
#define CHG_MPPT    (0x03)      //  MPPT mode
#define CMD_MASK    (0xfc)      // Command mask
#define CMD_LVP     (0x04)      //  LVP enabled
#define CMD_WDT     (0x08)      //  WDT enabled
#define CMD_PCY     (0x10)      //  PCY in progress
#define CMD_ENA     (0x20)      //  Enable charging
#define CMD_WDTACT  (0x40)      //  WDT actioned

#define LED_ON()    {LATAbits.LATA1  = 1;}
#define LED_OFF()   {LATAbits.LATA1  = 0;}
#define LED_INV()   {LATAbits.LATA1 ^= 1;}

#define DCDC_ON()   {LATBbits.LATB5  = 0;}
#define DCDC_OFF()  {LATBbits.LATB5  = 1;}

#define EEADR_VMPP  (0)
#define EEADR_VLIM  (1)
#define EEADR_ILIM  (2)
#define VMPP_DEF    (87)        // 8.7V (lsb=0.1)
#define VLIM_DEF    (210)       // 4.2V (lsb=0.02)
#define ILIM_DEF    (100)       // 1.0A (lsb=0.01)
__EEPROM_DATA (VMPP_DEF, VLIM_DEF, ILIM_DEF, 0, 0, 0, 0, 0);

#define REGS_SIZE   (8)         // I2C register size
unsigned char regs[REGS_SIZE];  // I2C register
#define ADDR_VBAT   (0)         // I2C Addr 0: Vbat (ro)
#define ADDR_IBAT   (1)         // I2C Addr 1: Ibat (ro)
#define ADDR_VPV    (2)         // I2C Addr 2: Vpv  (ro)
#define ADDR_CMD    (3)         // I2C Addr 3: cmd/stat (rw)
#define ADDR_VMPP   (4)         // I2C Addr 4: Vmpp (rw)
#define ADDR_VLIM   (5)         // I2C Addr 5: CV Limit (rw)
#define ADDR_ILIM   (6)         // I2C Addr 6: CC Limit (rw)
#define REG_VBAT    (regs[ADDR_VBAT])
#define REG_IBAT    (regs[ADDR_IBAT])
#define REG_VPV     (regs[ADDR_VPV])
#define REG_CMD     (regs[ADDR_CMD])
#define REG_VMPP    (regs[ADDR_VMPP])
#define REG_VLIM    (regs[ADDR_VLIM])
#define REG_ILIM    (regs[ADDR_ILIM])

unsigned char regs_addr;
unsigned char data_is_addr;
unsigned char dirty;
#define DIRTY_VMPP  (0x01)
#define DIRTY_VLIM  (0x02)
#define DIRTY_ILIM  (0x04)

unsigned char xtsk_10ms;
unsigned char xtsk_1s;
unsigned char count_10ms;

unsigned char i2cwdt;
#define I2CWDT_TOUT (30)

// Interrupt handler
void interrupt isr(void)
{
    unsigned char x;

    if(PIR1bits.TMR2IF){
        // 100Hz periodic timer interrupt
        PIR1bits.TMR2IF = 0;
        xtsk_10ms = 1;      // Command 10ms task
        if(count_10ms++ > 99){
            count_10ms = 0;
            xtsk_1s = 1;    // Command 1s task
        }
    }
    
    if(PIR1bits.SSP1IF){
	// I2C interrupt (Code refer to picfun.com)
        i2cwdt = 0; // Clear I2C watchdog
	if(SSP1STATbits.R_nW){
            // Master read request (slave send)
            if (SSP1STATbits.BF == 1){
                // After read addr
		x = SSP1BUF ;   // AhXf[^ǂ݂
		while(SSP1CON1bits.CKP | SSP1STATbits.BF);
                if(regs_addr < REGS_SIZE){
                    SSP1BUF = regs[regs_addr];   // Mf[^̃Zbg(1ڂ̃f[^)
                    regs_addr++;
                } else {
                    SSP1BUF = 0; // Dummy null
                }
                PIR1bits.SSP1IF = 0;         // ݃tONA
                SSP1CON1bits.CKP = 1 ;       // SCLCJ(ʐM̍ĊJ)
            } else {
                // f[^̑MACK󂯎ɂ銄荞݂Ɣf
		if (SSP1CON2bits.ACKSTAT==0) {
                    // }X^[ACKȂ玟̃f[^𑗐M
                    while( SSP1CON1bits.CKP | SSP1STATbits.BF);
                    if(regs_addr < REGS_SIZE){
                        SSP1BUF = regs[regs_addr];  // Mf[^̃Zbg(2ڈȍ~̃f[^)
                        regs_addr++ ;
                    } else {
                        SSP1BUF = 0; // Dummy null
                    }
                    PIR1bits.SSP1IF = 0 ;   // ݃tONA
                    SSP1CON1bits.CKP = 1 ;  // SCLCJ(ʐM̍ĊJ)
                } else {
                    // }X^[NOACKŉꂽ(MI)
                    PIR1bits.SSP1IF = 0 ;   // ݃tONA
                }
            }
        } else {
            // Master write request (slave recieve)
            if(SSP1STATbits.D_nA == 0){
                // I2C address
                x = SSP1BUF;    // Dummy read (i2c address)
                data_is_addr = 1;
            } else {
                // Data
                x = SSP1BUF;    // Data read
                if(data_is_addr){
                    // Set register address
                    data_is_addr = 0;
                    regs_addr = x;
                } else {
                    // Set register data
                    if(regs_addr < REGS_SIZE){
                        switch(regs_addr){
                            case ADDR_CMD:
                                REG_CMD = (x & CMD_MASK) | (REG_CMD & CHG_MASK);
                                break;
                            case ADDR_VMPP:
                                dirty |= DIRTY_VMPP;
                                REG_VMPP = x;
                                break;
                            case ADDR_VLIM:
                                dirty |= DIRTY_VLIM;
                                REG_VLIM = x;
                                break;
                            case ADDR_ILIM:
                                dirty |= DIRTY_ILIM;
                                REG_ILIM = x;
                                break;
                            default:
                                regs[regs_addr] = x;
                        }
                        regs_addr++;
                    }
                }
            }
            PIR1bits.SSP1IF = 0;    // ݃tONA
            SSP1CON1bits.CKP = 1;   // SCLCJ(ʐM̍ĊJ)
	}
    }
}

#define TS      (0.01F) // sampling time
// Create discrete 1st order lpf coeff using tustin z-conversion
//     wc
//  --------
//   S + wc
#define LPF10_WC (2.0F*3.14F*10.0F)  // 10Hz
#define LPF10_N0 ((double)(LPF10_WC*TS/(2.0F+LPF10_WC*TS)))     // num[0]
#define LPF10_N1 ((double)(LPF10_WC*TS/(2.0F+LPF10_WC*TS)))     // num[1]
#define LPF10_D1 ((double)((-2.0F+LPF10_WC*TS)/(2.0F+LPF10_WC*TS))) // den[1]

// Create discrete transfer function coeff using tustin z-conversion
//      1
//  ----------
//   S + TF_A
#define TF_A    (0.001F)
#define TF_N0   ((double)(TS/(2.0F+TF_A*TS)))                   // num[0]
#define TF_N1   ((double)(TS/(2.0F+TF_A*TS)))                   // num[1]
#define TF_D1   ((double)((-2.0F+TF_A*TS)/(2.0F+TF_A*TS)))      // den[1]

// Transfer function state buffer
struct tfbuf_st {
    double in_z1;   // Previous input (1/z)
    double out_z1;  // Previous output  (1/z)
} tfbuf[6];

// Initialize both input/output buffer at given value
void tfbuf_init(struct tfbuf_st *pbuf, double value)
{
    pbuf->in_z1 = value;
    pbuf->out_z1 = value;
}

// PI-controller
double pi_control(double obj, double curr, double pgain, double igain, struct tfbuf_st *pbuf)
{
    double err, pout, iin, iout;

    // error
    err = obj - curr;

    // P Control
    pout = err * pgain;
    
    // I Control (Not using pure-integration(1/s), using 1/(s+x) (s<<1) instead
    iin = err * igain;
    iout  = iin * TF_N0;
    iout += pbuf->in_z1 * TF_N1;
    iout -= pbuf->out_z1 * TF_D1;
    if(iout > 1.0) iout = 1.0;
    if(iout < 0.0) iout = 0.0;
    // Save for next time
    pbuf->in_z1 = iin;
    pbuf->out_z1 = iout;
    
    // Return P+I
    return (pout + iout);
}

// Low pass filter
double lpf10hz(double in, struct tfbuf_st *pbuf)
{
    double tmp;

    tmp  = in * LPF10_N0;
    tmp += pbuf->in_z1 * LPF10_N1;
    tmp -= pbuf->out_z1 * LPF10_D1;

    pbuf->in_z1 = in;
    pbuf->out_z1 = tmp;

    return tmp;
}

// A/D conversion
double adc_getvalue(unsigned char ch, double lsb)
{
    short adraw;
    double value;

    ADCON0bits.CHS = ch;        // Select channel
    __delay_us(5);              // Acquision time (>4.42us)

    ADCON0bits.GO_nDONE = 1;    // Start conversion
    while(ADCON0bits.GO_nDONE) continue;    // Wait until done

    adraw = (((short)(ADRESH & 0x03)) << 8) | (short)ADRESL;
    value = (double)adraw * lsb;

    return value;
}

// PWM output control
#define PWM_MAXWIDTH (1023)
#define PWM_MAXDUTY  (0.9F)
static short maxwidth;
void pwm_setduty(double duty)
{
    short width;
    if(duty < 0) duty = 0;
    if(duty > PWM_MAXDUTY) duty = PWM_MAXDUTY;
    width = (short)(duty * (double)PWM_MAXWIDTH);
    //if(pwm_value > PWM_MAXVALUE) pwm_value = PWM_MAXVALUE;    // PR1 = 127, 9bit resolution
    if(maxwidth < PWM_MAXWIDTH){
        maxwidth ++;
    }
    if(width > maxwidth) width = maxwidth;
    CCPR1L = (unsigned char) (width >> 2);              // Upper 8bit
    CCP1CONbits.DC1B = (unsigned char)(width & 0x03);   // Lower 2bit
}

void pwm_slowstart(void)
{
    maxwidth = 0;
}

// Set i2c readback data
void i2c_setdata(double vbat, double ibat, double vpv, unsigned char chgmode)
{
    short tmp;

    // Index0(V Batt) lsb=20mV
    tmp = (short) (vbat / I2CLSB_VBAT);
    if(tmp < 0) tmp = 0;
    if(tmp > 255) tmp = 255;
    REG_VBAT = (unsigned char)tmp;

    // Index1(I Chg) lsb=10mA
    tmp = (short) (ibat / I2CLSB_IBAT);
    if(tmp < 0) tmp = 0;
    if(tmp > 255) tmp = 255;
    REG_IBAT = (unsigned char)tmp;

    // Index2(V P.V.Cell) lsb=0.1V
    tmp = (short) (vpv / I2CLSB_VPV);
    if(tmp < 0) tmp = 0;
    if(tmp > 255) tmp = 255;
    REG_VPV = (unsigned char)tmp;

    // Index3(STAT) Update(lower 2bits), Keep CMD (upper 6bits)
    REG_CMD = (REG_CMD & CMD_MASK) | (chgmode & CHG_MASK);
}

// LED Blinker
#define BLINK_OFF   0
#define BLINK_NONE  1
#define BLINK_FAST  2
#define BLINK_SLOW  3
void blink_led(unsigned char blink)
{
    static unsigned char led_counter;
    switch(blink){
        default:
        case BLINK_OFF:
            LED_OFF();
            break;
        case BLINK_NONE:
            LED_ON();
            break;
        case BLINK_FAST:
            if(led_counter > 24){
                led_counter = 0;
                LED_INV();
            }
            break;
        case BLINK_SLOW:
            if(led_counter > 99){
                led_counter = 0;
                LED_INV();
            }
            break;
    }
    led_counter++;
}

// Read data from eeprom
unsigned char eeprom_read(unsigned char addr)
{
    EEADRL = addr;
    EECON1bits.CFGS = 0;
    EECON1bits.EEPGD =0;
    EECON1bits.RD = 1;
    return EEDATL;
}

// Write data to eeprom
void eeprom_write(unsigned char addr, unsigned char data)
{
    EEADRL = addr;
    EEDATL = data;
    EECON1bits.CFGS = 0;
    EECON1bits.EEPGD = 0;
    EECON1bits.WREN = 1;
    INTCONbits.GIE = 0;
    EECON2 = 0x55;
    EECON2 = 0xaa;
    EECON1bits.WR = 1;
    INTCONbits.GIE =1;
    EECON1bits.WREN = 0;
    while(EECON1bits.WR) continue;
}

// Main
void main(void)
{
    double vbatraw, vbat, ibatraw, ibat, vpvraw, vpv;
    double vcont, icont, mpcont, cont;
    unsigned char chgmode, blink, pcy_count = 0;
    double vmpp, vlim, ilim;

    // OSC Setup
    OSCCON = 0b01111010;    // No PLL, 16MHz Int Osc

    // Pin function control reg
    APFCON0 = 0b00000001;   // CCP1 on RB0

    // Port A Setup
    LATA = 0;
    ANSELA = 0b00000100;    // Use RA2 as analog (AN2:VPV input)
    TRISA =  0b00000100;    // Use RA2 as input others are output

    // Port B Setup
    LATB = 0;
    ANSELB = 0b00001100;    // Use RB3,2 as analog (AN9: VBAT, AN10:IBAT)
    TRISB  = 0b00011110;    // Use RB4,1 as I2c, RB3,2 as analog

    // Setup MSSP
    SSP1STAT= 0b10000000 ;      // Select 100kHz
    SSP1CON1= 0b00100110 ;      // Enable SDA SCL, 7bit slave mode
    SSP1CON2bits.SEN = 1 ;      // SCL(NbNXgb`)s
    SSP1ADD = I2C_ADDR << 1;    // }CAhX̐ݒ
    SSP1MSK = 0b11111110 ;      // AhXrp}XNf[^
    PIE1bits.SSP1IE = 1 ;       // SSP(I2C)荞݂
    PIR1bits.SSP1IF = 0;        // SSP(I2C)荞݃tONA
    PIE2bits.BCL1IE = 1 ;       // MSSP(I2C)oXՓˊ荞݂
    PIR2bits.BCL1IF = 0;        // MSSP(I2C)oXՓˊ荞݃tONA
    // INTCONbits.PEIE = 1;     // ӑu荞݂
    // INTCONbits.GIE = 1;      // S荞ݏ

    // Periodic timer (TMR2) Setup
    T2CON = 0b01001010;         // Prescale 1/16, Postscale 1/10
    PR2 = 250;                  // Set PR2 250 -> 1/16 * 1/10 * 1/250 => 1/40000
    T2CONbits.TMR2ON = 1;       // Enable TMR2
    PIE1bits.TMR2IE = 1;        // Enable TMR2 Int

    // Setup FVR , VGA
    FVRCON = 0b10000010;        // Enable FVR, ADC Ref output 2x (2.048V)
    while(FVRCONbits.FVRRDY == 0) continue; // Wait until FVR ready

    // Setup A/D
    ADCON1 = 0b11110011;        // Right justified, RC Osc, REF- = AVSS, REF+ = FVR
    ADCON0bits.ADON = 1;

    // Setup PWM
    // TRISBbits.TRISB0 = 0;    // RB0 output
    APFCON0bits.CCP1SEL = 1;    // CCP1 on RB0
    T4CONbits.T4CKPS = 0;       // TMR4 Prescaler 1/1
    PR4 = 255;                  // 15.6kHz 10bit resolution
    CCPTMRSbits.C1TSEL = 0b01;  // CCCP1 Based on TMR4
    CCP1CON = 0b00001100;       // ECCP Single out, Active-high
    pwm_setduty(0);             // Set output duty 0
    T4CONbits.TMR4ON = 1;       // Enable TMR

    // Set default command
    REG_CMD = CMD_LVP | CMD_ENA;

    // Read from eeprom
    REG_VMPP = eeprom_read(EEADR_VMPP);
    REG_VLIM = eeprom_read(EEADR_VLIM);
    REG_ILIM = eeprom_read(EEADR_ILIM);

    // Set control target
    vmpp = (double)REG_VMPP * I2CLSB_VPV;
    vlim = (double)REG_VLIM * I2CLSB_VBAT;
    ilim = (double)REG_ILIM * I2CLSB_IBAT;

    // Set watchdog 128ms
    WDTCONbits.SWDTEN = 1;      // Enable watchdog
    WDTCONbits.WDTPS = 0b00111; // Watch dog duration 128ms

    // Enable interrupt
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;

    while(1){

        // 1s task
        if(xtsk_1s){
            xtsk_1s = 0;
            if(REG_CMD & CMD_WDT){
                if(i2cwdt++ > I2CWDT_TOUT){
                    REG_CMD |= (CMD_PCY | CMD_WDTACT);
                    i2cwdt = 0;
                }
            }
        }

        // 10ms task
        if(xtsk_10ms){
            xtsk_10ms = 0;

            //LED_ON();

            // Changes to take effect and save to eeprom
            if(dirty & DIRTY_VMPP){
                dirty &= ~DIRTY_VMPP;
                eeprom_write(EEADR_VMPP, REG_VMPP);
                vmpp = (double)REG_VMPP * I2CLSB_VPV;
            }
            if(dirty & DIRTY_VLIM){
                dirty &= ~DIRTY_VLIM;
                eeprom_write(EEADR_VLIM, REG_VLIM);
                vlim = (double)REG_VLIM * I2CLSB_VBAT;
            }
            if(dirty & DIRTY_ILIM){
                dirty &= ~DIRTY_ILIM;
                eeprom_write(EEADR_ILIM, REG_ILIM);
                ilim = (double)REG_ILIM * I2CLSB_IBAT;
            }
            
            // A/D Conversion
            vbatraw = adc_getvalue(ADCH_VBAT, ADLSB_VBAT);  // V Battery
            ibatraw = adc_getvalue(ADCH_IBAT, ADLSB_IBAT);  // I Battery
            vpvraw  = adc_getvalue(ADCH_VPV,  ADLSB_VPV);   // V PV Cell

            // Low battery voltage protection (Prevent over discharge)
            if(REG_CMD & CMD_LVP){
                if(vbatraw < LVPOFF) DCDC_OFF();
                if(vbatraw > LVPON)  DCDC_ON();
            } else {
                DCDC_ON();
            }

            // Handle power-cycle (reboot)
            if(REG_CMD & CMD_PCY){
                if(pcy_count < 200){
                    pcy_count ++;
                    DCDC_OFF();
                } else {
                    REG_CMD &= ~CMD_PCY;
                    pcy_count = 0;
                    DCDC_ON();
                }
            }

            // Sleep control
            //if((vpvraw < VREADY) && (~REG_CMD & CMD_PCY) || (~REG_CMD & CMD_ENA)){
            if((~REG_CMD & CMD_PCY) && ((vpvraw < VREADY) || (~REG_CMD & CMD_ENA))){
                // Sleep state
                pwm_setduty(0);
                pwm_slowstart();
                blink_led(BLINK_OFF);
               
                // Set ibat to 0 due to current op-amp is not operating properly
                i2c_setdata(vbatraw, 0, vpvraw, CHG_OFF);

                ADCON0bits.ADON = 0;     // Disable A/D
                FVRCONbits.FVREN = 0;    // Disable FVR

                WDTCONbits.WDTPS = 0b01010; // Watch dog duration 1s (alarm)
                SLEEP();
                WDTCONbits.WDTPS = 0b00111; // Watch dog duration 128ms (normal)

                ADCON0bits.ADON = 1;     // Enable A/D
                FVRCONbits.FVREN = 1;    // Enable FVR
                while(FVRCONbits.FVRRDY == 0) continue; // Wait until FVR ready

                xtsk_10ms = 1;
                xtsk_1s = 1;

                tfbuf_init(&tfbuf[0], vbatraw);
                tfbuf_init(&tfbuf[1], ibatraw);
                tfbuf_init(&tfbuf[2], vpvraw);
                tfbuf_init(&tfbuf[3], 0.0F);
                tfbuf_init(&tfbuf[4], 0.0F);
                tfbuf_init(&tfbuf[5], 0.0F);
                continue;
            }

            // Low pass filter
            vbat = lpf10hz(vbatraw, &tfbuf[0]);
            ibat = lpf10hz(ibatraw, &tfbuf[1]);
            vpv  = lpf10hz(vpvraw,  &tfbuf[2]);

            // Safety guard at just in front of error control
            if(vlim > VLIM_MAX) vlim = VLIM_MAX;
            if(ilim > ILIM_MAX) ilim = ILIM_MAX;

            // PI control
            vcont  = pi_control(vlim, vbat, PVGAIN, IVGAIN, &tfbuf[3]);
            icont  = pi_control(ilim, ibat, PIGAIN, IIGAIN, &tfbuf[4]);
            mpcont = pi_control(vpv,  vmpp, PMGAIN, IMGAIN, &tfbuf[5]);

            // Mode select
            // MPPT
            cont = mpcont;
            chgmode = CHG_MPPT;
            blink = BLINK_FAST;

            // CV override
            if(vcont < cont){
                cont = vcont;
                chgmode = CHG_CV;
                blink = BLINK_NONE;
            }

            // CC override
            if(icont < cont){
                cont = icont;
                chgmode = CHG_CC;
                blink = BLINK_SLOW;
            }

            // Output
            pwm_setduty(cont);
            i2c_setdata(vbat, ibat, vpv, chgmode);
            blink_led(blink);

            //LED_OFF();

            CLRWDT();
        }
    }
}
// EOF